@@ -7,7 +7,9 @@
 #include <stan/math/rev/functor/ode_store_sensitivities.hpp>
 #include <stan/math/prim/err.hpp>
 #include <stan/math/prim/fun/value_of.hpp>
+#include <stan/math/prim/functor/apply.hpp>
 #include <stan/math/prim/functor/for_each.hpp>
+#include <sundials/sundials_context.h>
 #include <cvodes/cvodes.h>
 #include <nvector/nvector_serial.h>
 #include <sunmatrix/sunmatrix_dense.h>
@@ -46,29 +48,21 @@ class cvodes_integrator_adjoint_vari : public vari_base {
   static constexpr bool is_var_only_ts_{
       is_var_ts_ && !(is_var_t0_ || is_var_y0_t0_ || is_any_var_args_)};
 
-  arena_t<std::vector<Eigen::VectorXd>> y_;
-  arena_t<std::vector<T_ts>> ts_;
-  arena_t<Eigen::Matrix<T_y0_t0, Eigen::Dynamic, 1>> y0_;
-  arena_t<Eigen::VectorXd> absolute_tolerance_forward_;
-  arena_t<Eigen::VectorXd> absolute_tolerance_backward_;
-  arena_t<Eigen::VectorXd> state_forward_;
-  arena_t<Eigen::VectorXd> state_backward_;
-  size_t num_args_vars_;
-  arena_t<Eigen::VectorXd> quad_;
-  arena_t<T_t0> t0_;
+  const size_t num_args_vars_;
 
-  double relative_tolerance_forward_;
-  double relative_tolerance_backward_;
-  double relative_tolerance_quadrature_;
-  double absolute_tolerance_quadrature_;
-  long int max_num_steps_;                  // NOLINT(runtime/int)
-  long int num_steps_between_checkpoints_;  // NOLINT(runtime/int)
-  size_t N_;
+  const double relative_tolerance_forward_;
+  const double relative_tolerance_backward_;
+  const double relative_tolerance_quadrature_;
+  const double absolute_tolerance_quadrature_;
+  const long int max_num_steps_;                  // NOLINT(runtime/int)
+  const long int num_steps_between_checkpoints_;  // NOLINT(runtime/int)
+  const size_t N_;
   std::ostream* msgs_;
+  vari** y_return_varis_;
   vari** args_varis_;
-  int interpolation_polynomial_;
-  int solver_forward_;
-  int solver_backward_;
+  const int interpolation_polynomial_;
+  const int solver_forward_;
+  const int solver_backward_;
   int index_backward_;
   bool backward_is_initialized_{false};
 
@@ -78,9 +72,21 @@ class cvodes_integrator_adjoint_vari : public vari_base {
    * vari class).
    */
   struct cvodes_solver : public chainable_alloc {
+    sundials::Context sundials_context_;
     const std::string function_name_str_;
     const std::decay_t<F> f_;
     const size_t N_;
+    std::vector<Eigen::VectorXd> y_;
+
+    std::vector<T_ts> ts_;
+    Eigen::Matrix<T_y0_t0, Eigen::Dynamic, 1> y0_;
+    Eigen::VectorXd absolute_tolerance_forward_;
+    Eigen::VectorXd absolute_tolerance_backward_;
+    Eigen::VectorXd state_forward_;
+    Eigen::VectorXd state_backward_;
+    Eigen::VectorXd quad_;
+    T_t0 t0_;
+
     N_Vector nv_state_forward_;
     N_Vector nv_state_backward_;
     N_Vector nv_quad_;
@@ -91,40 +97,51 @@ class cvodes_integrator_adjoint_vari : public vari_base {
     SUNMatrix A_backward_;
     SUNLinearSolver LS_backward_;
     void* cvodes_mem_;
-    std::vector<Eigen::Matrix<T_Return, Eigen::Dynamic, 1>> y_return_;
     std::tuple<T_Args...> local_args_tuple_;
     const std::tuple<
         promote_scalar_t<partials_type_t<scalar_type_t<T_Args>>, T_Args>...>
         value_of_args_tuple_;
 
-    template <typename FF, typename StateFwd, typename StateBwd, typename Quad,
-              typename AbsTolFwd, typename AbsTolBwd>
-    cvodes_solver(const char* function_name, FF&& f, size_t N,
-                  size_t num_args_vars, size_t ts_size, int solver_forward,
-                  StateFwd& state_forward, StateBwd& state_backward, Quad& quad,
-                  AbsTolFwd& absolute_tolerance_forward,
-                  AbsTolBwd& absolute_tolerance_backward, const T_Args&... args)
+    template <typename FF>
+    cvodes_solver(const char* function_name, FF&& f, size_t N, const T_y0& y0,
+                  const T_t0& t0, const std::vector<T_ts>& ts,
+                  const Eigen::VectorXd& absolute_tolerance_forward,
+                  const Eigen::VectorXd& absolute_tolerance_backward,
+                  size_t num_args_vars, int solver_forward,
+                  const T_Args&... args)
         : chainable_alloc(),
-          f_(std::forward<FF>(f)),
+          sundials_context_(),
           function_name_str_(function_name),
-          y_return_(ts_size),
-          nv_state_forward_(N_VMake_Serial(N, state_forward.data())),
-          nv_state_backward_(N_VMake_Serial(N, state_backward.data())),
-          nv_quad_(N_VMake_Serial(num_args_vars, quad.data())),
-          nv_absolute_tolerance_forward_(
-              N_VMake_Serial(N, absolute_tolerance_forward.data())),
-          nv_absolute_tolerance_backward_(
-              N_VMake_Serial(N, absolute_tolerance_backward.data())),
-          A_forward_(SUNDenseMatrix(N, N)),
-          A_backward_(SUNDenseMatrix(N, N)),
-          LS_forward_(
-              N == 0 ? nullptr
-                     : SUNDenseLinearSolver(nv_state_forward_, A_forward_)),
-          LS_backward_(
-              N == 0 ? nullptr
-                     : SUNDenseLinearSolver(nv_state_backward_, A_backward_)),
+          f_(std::forward<FF>(f)),
           N_(N),
-          cvodes_mem_(CVodeCreate(solver_forward)),
+          y_(ts.size()),
+          ts_(ts.begin(), ts.end()),
+          y0_(y0),
+          absolute_tolerance_forward_(absolute_tolerance_forward),
+          absolute_tolerance_backward_(absolute_tolerance_backward),
+          state_forward_(value_of(y0)),
+          state_backward_(Eigen::VectorXd::Zero(N)),
+          quad_(Eigen::VectorXd::Zero(num_args_vars)),
+          t0_(t0),
+          nv_state_forward_(
+              N_VMake_Serial(N, state_forward_.data(), sundials_context_)),
+          nv_state_backward_(
+              N_VMake_Serial(N, state_backward_.data(), sundials_context_)),
+          nv_quad_(
+              N_VMake_Serial(num_args_vars, quad_.data(), sundials_context_)),
+          nv_absolute_tolerance_forward_(N_VMake_Serial(
+              N, absolute_tolerance_forward_.data(), sundials_context_)),
+          nv_absolute_tolerance_backward_(N_VMake_Serial(
+              N, absolute_tolerance_backward_.data(), sundials_context_)),
+          A_forward_(SUNDenseMatrix(N, N, sundials_context_)),
+          A_backward_(SUNDenseMatrix(N, N, sundials_context_)),
+          LS_forward_(N == 0 ? nullptr
+                             : SUNLinSol_Dense(nv_state_forward_, A_forward_,
+                                               sundials_context_)),
+          LS_backward_(N == 0 ? nullptr
+                              : SUNLinSol_Dense(nv_state_backward_, A_backward_,
+                                                sundials_context_)),
+          cvodes_mem_(CVodeCreate(solver_forward, sundials_context_)),
           local_args_tuple_(deep_copy_vars(args)...),
           value_of_args_tuple_(value_of(args)...) {
       if (cvodes_mem_ == nullptr) {
@@ -204,16 +221,7 @@ class cvodes_integrator_adjoint_vari : public vari_base {
       int interpolation_polynomial, int solver_forward, int solver_backward,
       std::ostream* msgs, const T_Args&... args)
       : vari_base(),
-        y_(ts.size()),
-        ts_(ts.begin(), ts.end()),
-        y0_(y0),
-        absolute_tolerance_forward_(absolute_tolerance_forward),
-        absolute_tolerance_backward_(absolute_tolerance_backward),
-        state_forward_(value_of(y0)),
-        state_backward_(y0.size()),
         num_args_vars_(count_vars(args...)),
-        quad_(num_args_vars_),
-        t0_(t0),
         relative_tolerance_forward_(relative_tolerance_forward),
         relative_tolerance_backward_(relative_tolerance_backward),
         relative_tolerance_quadrature_(relative_tolerance_quadrature),
@@ -222,6 +230,9 @@ class cvodes_integrator_adjoint_vari : public vari_base {
         num_steps_between_checkpoints_(num_steps_between_checkpoints),
         N_(y0.size()),
         msgs_(msgs),
+        y_return_varis_(is_var_return_ ? ChainableStack::instance_->memalloc_
+                                             .alloc_array<vari*>(N_ * ts.size())
+                                       : nullptr),
         args_varis_([&args..., num_vars = this->num_args_vars_]() {
           vari** vari_mem
               = ChainableStack::instance_->memalloc_.alloc_array<vari*>(
@@ -245,15 +256,15 @@ class cvodes_integrator_adjoint_vari : public vari_base {
     check_positive_finite(function_name, "relative_tolerance_forward",
                           relative_tolerance_forward_);
     check_positive_finite(function_name, "absolute_tolerance_forward",
-                          absolute_tolerance_forward_);
+                          absolute_tolerance_forward);
     check_size_match(function_name, "absolute_tolerance_forward",
-                     absolute_tolerance_forward_.size(), "states", N_);
+                     absolute_tolerance_forward.size(), "states", N_);
     check_positive_finite(function_name, "relative_tolerance_backward",
                           relative_tolerance_backward_);
     check_positive_finite(function_name, "absolute_tolerance_backward",
-                          absolute_tolerance_backward_);
+                          absolute_tolerance_backward);
     check_size_match(function_name, "absolute_tolerance_backward",
-                     absolute_tolerance_backward_.size(), "states", N_);
+                     absolute_tolerance_backward.size(), "states", N_);
     check_positive_finite(function_name, "relative_tolerance_quadrature",
                           relative_tolerance_quadrature_);
     check_positive_finite(function_name, "absolute_tolerance_quadrature",
@@ -275,10 +286,10 @@ class cvodes_integrator_adjoint_vari : public vari_base {
       invalid_argument(function_name, "solver_backward", solver_backward_, "",
                        ", must be 1 for Adams or 2 for BDF backward solver");
 
-    solver_ = new cvodes_solver(
-        function_name, f, N_, num_args_vars_, ts_.size(), solver_forward_,
-        state_forward_, state_backward_, quad_, absolute_tolerance_forward_,
-        absolute_tolerance_backward_, args...);
+    solver_ = new cvodes_solver(function_name, std::forward<FF>(f), N_, y0, t0,
+                                ts, absolute_tolerance_forward,
+                                absolute_tolerance_backward, num_args_vars_,
+                                solver_forward_, args...);
 
     stan::math::for_each(
         [func_name = function_name](auto&& arg) {
@@ -286,80 +297,62 @@ class cvodes_integrator_adjoint_vari : public vari_base {
         },
         solver_->local_args_tuple_);
 
-    check_flag_sundials(
+    CHECK_CVODES_CALL(
         CVodeInit(solver_->cvodes_mem_, &cvodes_integrator_adjoint_vari::cv_rhs,
-                  value_of(t0_), solver_->nv_state_forward_),
-        "CVodeInit");
+                  value_of(solver_->t0_), solver_->nv_state_forward_));
 
     // Assign pointer to this as user data
-    check_flag_sundials(
-        CVodeSetUserData(solver_->cvodes_mem_, reinterpret_cast<void*>(this)),
-        "CVodeSetUserData");
+    CHECK_CVODES_CALL(
+        CVodeSetUserData(solver_->cvodes_mem_, reinterpret_cast<void*>(this)));
 
     cvodes_set_options(solver_->cvodes_mem_, max_num_steps_);
 
-    check_flag_sundials(
+    CHECK_CVODES_CALL(
         CVodeSVtolerances(solver_->cvodes_mem_, relative_tolerance_forward_,
-                          solver_->nv_absolute_tolerance_forward_),
-        "CVodeSVtolerances");
+                          solver_->nv_absolute_tolerance_forward_));
 
-    check_flag_sundials(
-        CVodeSetLinearSolver(solver_->cvodes_mem_, solver_->LS_forward_,
-                             solver_->A_forward_),
-        "CVodeSetLinearSolver");
+    CHECK_CVODES_CALL(CVodeSetLinearSolver(
+        solver_->cvodes_mem_, solver_->LS_forward_, solver_->A_forward_));
 
-    check_flag_sundials(
+    CHECK_CVODES_CALL(
         CVodeSetJacFn(solver_->cvodes_mem_,
-                      &cvodes_integrator_adjoint_vari::cv_jacobian_rhs_states),
-        "CVodeSetJacFn");
+                      &cvodes_integrator_adjoint_vari::cv_jacobian_rhs_states));
 
     // initialize backward sensitivity system of CVODES as needed
     if (is_var_return_ && !is_var_only_ts_) {
-      check_flag_sundials(
-          CVodeAdjInit(solver_->cvodes_mem_, num_steps_between_checkpoints_,
-                       interpolation_polynomial_),
-          "CVodeAdjInit");
+      CHECK_CVODES_CALL(CVodeAdjInit(solver_->cvodes_mem_,
+                                     num_steps_between_checkpoints_,
+                                     interpolation_polynomial_));
     }
 
     /**
      * Solve the ODE initial value problem y' = f(t, y), y(t0) = y0 at a set of
      * times, { t1, t2, t3, ... } using the requested forward solver of CVODES.
      */
-    const auto ts_dbl = value_of(ts_);
+    const auto ts_dbl = value_of(solver_->ts_);
 
-    double t_init = value_of(t0_);
+    double t_init = value_of(solver_->t0_);
     for (size_t n = 0; n < ts_dbl.size(); ++n) {
       double t_final = ts_dbl[n];
       if (t_final != t_init) {
         if (is_var_return_ && !is_var_only_ts_) {
           int ncheck;
 
-          int error_code
-              = CVodeF(solver_->cvodes_mem_, t_final,
-                       solver_->nv_state_forward_, &t_init, CV_NORMAL, &ncheck);
-
-          if (unlikely(error_code == CV_TOO_MUCH_WORK)) {
-            throw_domain_error(solver_->function_name_str_.c_str(), "", t_final,
-                               "Failed to integrate to next output time (",
-                               ") in less than max_num_steps steps");
-          } else {
-            check_flag_sundials(error_code, "CVodeF");
-          }
+          CHECK_CVODES_CALL(CVodeF(solver_->cvodes_mem_, t_final,
+                                   solver_->nv_state_forward_, &t_init,
+                                   CV_NORMAL, &ncheck));
         } else {
-          int error_code
-              = CVode(solver_->cvodes_mem_, t_final, solver_->nv_state_forward_,
-                      &t_init, CV_NORMAL);
-
-          if (unlikely(error_code == CV_TOO_MUCH_WORK)) {
-            throw_domain_error(solver_->function_name_str_.c_str(), "", t_final,
-                               "Failed to integrate to next output time (",
-                               ") in less than max_num_steps steps");
-          } else {
-            check_flag_sundials(error_code, "CVode");
-          }
+          CHECK_CVODES_CALL(CVode(solver_->cvodes_mem_, t_final,
+                                  solver_->nv_state_forward_, &t_init,
+                                  CV_NORMAL));
         }
       }
-      store_state(n, state_forward_, solver_->y_return_[n]);
+      solver_->y_[n] = solver_->state_forward_;
+      if (is_var_return_) {
+        for (std::size_t i = 0; i < N_; ++i)
+          y_return_varis_[N_ * n + i]
+              = new vari(solver_->state_forward_.coeff(i), false);
+      }
 
       t_init = t_final;
     }
@@ -374,16 +367,14 @@ class cvodes_integrator_adjoint_vari : public vari_base {
    */
   void store_state(std::size_t n, const Eigen::VectorXd& state,
                    Eigen::Matrix<var, Eigen::Dynamic, 1>& state_return) {
-    y_[n] = state;
     state_return.resize(N_);
     for (size_t i = 0; i < N_; i++) {
-      state_return.coeffRef(i) = var(new vari(state.coeff(i), false));
+      state_return.coeffRef(i) = var(y_return_varis_[N_ * n + i]);
     }
   }
 
   void store_state(std::size_t n, const Eigen::VectorXd& state,
                    Eigen::Matrix<double, Eigen::Dynamic, 1>& state_return) {
-    y_[n] = state;
     state_return = state;
   }
 
@@ -395,7 +386,11 @@ class cvodes_integrator_adjoint_vari : public vari_base {
    *   solution time (excluding the initial state)
    */
   std::vector<Eigen::Matrix<T_Return, Eigen::Dynamic, 1>> solution() noexcept {
-    return solver_->y_return_;
+    std::vector<Eigen::Matrix<T_Return, Eigen::Dynamic, 1>> y_return(
+        solver_->ts_.size());
+    for (std::size_t n = 0; n < solver_->ts_.size(); ++n)
+      store_state(n, solver_->y_[n], y_return[n]);
+    return y_return;
   }
 
   /**
@@ -412,14 +407,14 @@ class cvodes_integrator_adjoint_vari : public vari_base {
     // integration
     if (is_var_ts_) {
       Eigen::VectorXd step_sens = Eigen::VectorXd::Zero(N_);
-      for (int i = 0; i < ts_.size(); ++i) {
+      for (int i = 0; i < solver_->ts_.size(); ++i) {
         for (int j = 0; j < N_; ++j) {
-          step_sens.coeffRef(j)
-              += forward_as<var>(solver_->y_return_[i].coeff(j)).adj();
+          step_sens.coeffRef(j) += y_return_varis_[N_ * i + j]->adj_;
         }
 
-        adjoint_of(ts_[i]) += step_sens.dot(
-            rhs(value_of(ts_[i]), y_[i], solver_->value_of_args_tuple_));
+        adjoint_of(solver_->ts_[i])
+            += step_sens.dot(rhs(value_of(solver_->ts_[i]), solver_->y_[i],
+                                 solver_->value_of_args_tuple_));
         step_sens.setZero();
       }
 
@@ -428,141 +423,117 @@ class cvodes_integrator_adjoint_vari : public vari_base {
       }
     }
 
-    state_backward_.setZero();
-    quad_.setZero();
+    solver_->state_backward_.setZero();
+    solver_->quad_.setZero();
 
     // At every time step, collect the adjoints from the output
     // variables and re-initialize the solver
-    double t_init = value_of(ts_.back());
-    for (int i = ts_.size() - 1; i >= 0; --i) {
+    double t_init = value_of(solver_->ts_.back());
+    for (int i = solver_->ts_.size() - 1; i >= 0; --i) {
       // Take in the adjoints from all the output variables at this point
       // in time
       for (int j = 0; j < N_; ++j) {
-        state_backward_.coeffRef(j)
-            += forward_as<var>(solver_->y_return_[i].coeff(j)).adj();
+        solver_->state_backward_.coeffRef(j)
+            += y_return_varis_[N_ * i + j]->adj_;
       }
 
-      double t_final = value_of((i > 0) ? ts_[i - 1] : t0_);
+      double t_final = value_of((i > 0) ? solver_->ts_[i - 1] : solver_->t0_);
       if (t_final != t_init) {
         if (unlikely(!backward_is_initialized_)) {
-          check_flag_sundials(CVodeCreateB(solver_->cvodes_mem_,
-                                           solver_backward_, &index_backward_),
-                              "CVodeCreateB");
+          CHECK_CVODES_CALL(CVodeCreateB(solver_->cvodes_mem_, solver_backward_,
+                                         &index_backward_));
 
-          check_flag_sundials(
-              CVodeSetUserDataB(solver_->cvodes_mem_, index_backward_,
-                                reinterpret_cast<void*>(this)),
-              "CVodeSetUserDataB");
+          CHECK_CVODES_CALL(CVodeSetUserDataB(solver_->cvodes_mem_,
+                                              index_backward_,
+                                              reinterpret_cast<void*>(this)));
 
           // initialize CVODES backward machinery.
           // the states of the backward problem *are* the adjoints
           // of the ode states
-          check_flag_sundials(
+          CHECK_CVODES_CALL(
               CVodeInitB(solver_->cvodes_mem_, index_backward_,
                          &cvodes_integrator_adjoint_vari::cv_rhs_adj, t_init,
-                         solver_->nv_state_backward_),
-              "CVodeInitB");
+                         solver_->nv_state_backward_));
 
-          check_flag_sundials(
+          CHECK_CVODES_CALL(
               CVodeSVtolerancesB(solver_->cvodes_mem_, index_backward_,
                                  relative_tolerance_backward_,
-                                 solver_->nv_absolute_tolerance_backward_),
-              "CVodeSVtolerancesB");
+                                 solver_->nv_absolute_tolerance_backward_));
 
-          check_flag_sundials(
-              CVodeSetMaxNumStepsB(solver_->cvodes_mem_, index_backward_,
-                                   max_num_steps_),
-              "CVodeSetMaxNumStepsB");
+          CHECK_CVODES_CALL(CVodeSetMaxNumStepsB(
+              solver_->cvodes_mem_, index_backward_, max_num_steps_));
 
-          check_flag_sundials(CVodeSetLinearSolverB(
-                                  solver_->cvodes_mem_, index_backward_,
-                                  solver_->LS_backward_, solver_->A_backward_),
-                              "CVodeSetLinearSolverB");
+          CHECK_CVODES_CALL(CVodeSetLinearSolverB(
+              solver_->cvodes_mem_, index_backward_, solver_->LS_backward_,
+              solver_->A_backward_));
 
-          check_flag_sundials(
-              CVodeSetJacFnB(
-                  solver_->cvodes_mem_, index_backward_,
-                  &cvodes_integrator_adjoint_vari::cv_jacobian_rhs_adj_states),
-              "CVodeSetJacFnB");
+          CHECK_CVODES_CALL(CVodeSetJacFnB(
+              solver_->cvodes_mem_, index_backward_,
+              &cvodes_integrator_adjoint_vari::cv_jacobian_rhs_adj_states));
 
           // Allocate space for backwards quadrature needed when
           // parameters vary.
           if (is_any_var_args_) {
-            check_flag_sundials(
+            CHECK_CVODES_CALL(
                 CVodeQuadInitB(solver_->cvodes_mem_, index_backward_,
                                &cvodes_integrator_adjoint_vari::cv_quad_rhs_adj,
-                               solver_->nv_quad_),
-                "CVodeQuadInitB");
+                               solver_->nv_quad_));
 
-            check_flag_sundials(
+            CHECK_CVODES_CALL(
                 CVodeQuadSStolerancesB(solver_->cvodes_mem_, index_backward_,
                                        relative_tolerance_quadrature_,
-                                       absolute_tolerance_quadrature_),
-                "CVodeQuadSStolerancesB");
+                                       absolute_tolerance_quadrature_));
 
-            check_flag_sundials(CVodeSetQuadErrConB(solver_->cvodes_mem_,
-                                                    index_backward_, SUNTRUE),
-                                "CVodeSetQuadErrConB");
+            CHECK_CVODES_CALL(CVodeSetQuadErrConB(solver_->cvodes_mem_,
+                                                  index_backward_, SUNTRUE));
           }
 
           backward_is_initialized_ = true;
         } else {
           // just re-initialize the solver
-          check_flag_sundials(
-              CVodeReInitB(solver_->cvodes_mem_, index_backward_, t_init,
-                           solver_->nv_state_backward_),
-              "CVodeReInitB");
+
+          CHECK_CVODES_CALL(CVodeReInitB(solver_->cvodes_mem_, index_backward_,
+                                         t_init, solver_->nv_state_backward_));
 
           if (is_any_var_args_) {
-            check_flag_sundials(
-                CVodeQuadReInitB(solver_->cvodes_mem_, index_backward_,
-                                 solver_->nv_quad_),
-                "CVodeQuadReInitB");
+            CHECK_CVODES_CALL(CVodeQuadReInitB(
+                solver_->cvodes_mem_, index_backward_, solver_->nv_quad_));
           }
         }
 
-        int error_code = CVodeB(solver_->cvodes_mem_, t_final, CV_NORMAL);
-
-        if (unlikely(error_code == CV_TOO_MUCH_WORK)) {
-          throw_domain_error(solver_->function_name_str_.c_str(), "", t_final,
-                             "Failed to integrate backward to output time (",
-                             ") in less than max_num_steps steps");
-        } else {
-          check_flag_sundials(error_code, "CVodeB");
-        }
+        CHECK_CVODES_CALL(CVodeB(solver_->cvodes_mem_, t_final, CV_NORMAL));
 
         // obtain adjoint states and update t_init to time point
         // reached of t_final
-        check_flag_sundials(CVodeGetB(solver_->cvodes_mem_, index_backward_,
-                                      &t_init, solver_->nv_state_backward_),
-                            "CVodeGetB");
+        CHECK_CVODES_CALL(CVodeGetB(solver_->cvodes_mem_, index_backward_,
+                                    &t_init, solver_->nv_state_backward_));
 
         if (is_any_var_args_) {
-          check_flag_sundials(
-              CVodeGetQuadB(solver_->cvodes_mem_, index_backward_, &t_init,
-                            solver_->nv_quad_),
-              "CVodeGetQuadB");
+          CHECK_CVODES_CALL(CVodeGetQuadB(solver_->cvodes_mem_, index_backward_,
+                                          &t_init, solver_->nv_quad_));
         }
       }
     }
 
-    if (is_var_t0_) {
-      adjoint_of(t0_) += -state_backward_.dot(
-          rhs(t_init, value_of(y0_), solver_->value_of_args_tuple_));
-    }
-
     // After integrating all the way back to t0, we finally have the
     // the adjoints we wanted
+
+    // This is the dlog_density / d(initial_time_point) adjoint
+    if (is_var_t0_) {
+      adjoint_of(solver_->t0_) += -solver_->state_backward_.dot(
+          rhs(t_init, value_of(solver_->y0_), solver_->value_of_args_tuple_));
+    }
+
     // These are the dlog_density / d(initial_conditions[s]) adjoints
     if (is_var_y0_t0_) {
-      forward_as<arena_t<Eigen::Matrix<var, Eigen::Dynamic, 1>>>(y0_).adj()
-          += state_backward_;
+      forward_as<Eigen::Matrix<var, Eigen::Dynamic, 1>>(solver_->y0_).adj()
+          += solver_->state_backward_;
     }
 
     // These are the dlog_density / d(parameters[s]) adjoints
     if (is_any_var_args_) {
       for (size_t s = 0; s < num_args_vars_; ++s) {
-        args_varis_[s]->adj_ += quad_.coeff(s);
+        args_varis_[s]->adj_ += solver_->quad_.coeff(s);
       }
     }
   }
@@ -574,7 +545,7 @@ class cvodes_integrator_adjoint_vari : public vari_base {
   template <typename yT, typename... ArgsT>
   constexpr auto rhs(double t, const yT& y,
                      const std::tuple<ArgsT...>& args_tuple) const {
-    return apply(
+    return math::apply(
         [&](auto&&... args) { return solver_->f_(t, y, msgs_, args...); },
         args_tuple);
   }
@@ -664,13 +635,14 @@ class cvodes_integrator_adjoint_vari : public vari_base {
     // separately
     stan::math::for_each([](auto&& arg) { zero_adjoints(arg); },
                          solver_->local_args_tuple_);
+
     Eigen::Matrix<var, Eigen::Dynamic, 1> f_y_t_vars
         = rhs(t, y_vec, solver_->local_args_tuple_);
     check_size_match(solver_->function_name_str_.c_str(), "dy_dt",
                      f_y_t_vars.size(), "states", N_);
     f_y_t_vars.adj() = -Eigen::Map<Eigen::VectorXd>(NV_DATA_S(yB), N_);
     grad();
-    apply(
+    math::apply(
         [&qBdot](auto&&... args) {
           accumulate_adjoints(NV_DATA_S(qBdot), args...);
         },
